from pytgcalls.dispatcher.action import Action
from pytgcalls.dispatcher.dispatcher import Dispatcher
from pytgcalls.dispatcher.dispatcher_mixin import DispatcherMixin

__all__ = [
    'Action',
    'Dispatcher',
    'DispatcherMixin',
]



from pytgcalls.mtproto.data import WrapperBase


class GroupCallWrapper(WrapperBase):
    def __init__(self, call_id: int, params):
        self.id = call_id
        self.params = params  # its a DataJSON object with params.data attr



from .update_group_call_participants_wrapper import UpdateGroupCallParticipantsWrapper
from .update_group_call_wrapper import UpdateGroupCallWrapper

__all__ = [
    'UpdateGroupCallParticipantsWrapper',
    'UpdateGroupCallWrapper',
]



import asyncio
import logging
from typing import Callable, List

from typing import TYPE_CHECKING

from ..exceptions import PytgcallsError

if TYPE_CHECKING:
    from . import GroupCallNative

logger = logging.getLogger(__name__)


class Dispatcher:
    def __init__(self, available_actions: type):
        self.actions = available_actions
        self.__action_to_handlers = self.__build_handler_storage()

    def __build_handler_storage(self):
        logger.debug('Build storage of handlers for dispatcher.')
        return {action: [] for action in dir(self.actions) if not action.startswith('_')}

    def add_handler(self, callback: Callable, action: str) -> Callable:
        logger.debug(f'Add handler to {action} action...')
        if not asyncio.iscoroutinefunction(callback):
            raise PytgcallsError('Sync callback does not supported')

        try:
            handlers = self.__action_to_handlers[action]
            if callback in handlers:
                logger.debug('Handler is already set.')
                return callback

            handlers.append(callback)
        except KeyError:
            raise PytgcallsError('Invalid action')

        logger.debug('Handler added.')
        return callback

    def remove_handler(self, callback: Callable, action: str) -> bool:
        logger.debug(f'Remove handler of {action} action...')
        try:
            handlers = self.__action_to_handlers[action]
            for i in range(len(handlers)):
                if handlers[i] == callback:
                    del handlers[i]
                    return True
        except KeyError:
            raise PytgcallsError('Invalid action')

        return False

    def remove_all(self):
        self.__action_to_handlers = self.__build_handler_storage()

    def get_handlers(self, action: str) -> List[Callable]:
        try:
            logger.debug(f'Get {action} handlers...')
            return self.__action_to_handlers[action]
        except KeyError:
            raise PytgcallsError('Invalid action')

    def trigger_handlers(self, action: str, instance: 'GroupCallNative', *args, **kwargs):
        logger.debug(f'Trigger {action} handlers...')

        for handler in self.get_handlers(action):
            logger.debug(f'Trigger {handler.__name__}...')
            instance.get_event_loop().create_task(handler(instance, *args, **kwargs))



class PytgcallsBaseException(Exception):
    ...


class PytgcallsError(PytgcallsBaseException):
    ...


class CallBeforeStartError(PytgcallsBaseException):
    ...


class NotConnectedError(PytgcallsBaseException):
    ...


class GroupCallNotFoundError(PytgcallsBaseException):
    ...



from typing import Callable

import tgcalls
from pytgcalls.implementation import GroupCallBase, GroupCallBaseAction, GroupCallBaseDispatcherMixin
from pytgcalls.dispatcher import Action


class GroupCallFileAction(GroupCallBaseAction):
    PLAYOUT_ENDED = Action()
    '''When a input file is ended.'''


class GroupCallFileDispatcherMixin(GroupCallBaseDispatcherMixin):
    def on_playout_ended(self, func: Callable) -> Callable:
        """When a input file is ended.

        Args:
            func (`Callable`): A functions that accept group_call and filename args.

        Returns:
            `Callable`: passed to args callback function.
        """

        return self.add_handler(func, GroupCallFileAction.PLAYOUT_ENDED)


class GroupCallFile(GroupCallBase, GroupCallFileDispatcherMixin):
    def __init__(
        self,
        mtproto_bridge,
        input_filename: str = None,
        output_filename: str = None,
        play_on_repeat=True,
        enable_logs_to_console=False,
        path_to_log_file=None,
        outgoing_audio_bitrate_kbit=128,
    ):
        super().__init__(mtproto_bridge, enable_logs_to_console, path_to_log_file, outgoing_audio_bitrate_kbit)
        super(GroupCallFileDispatcherMixin, self).__init__(GroupCallFileAction)

        self.play_on_repeat = play_on_repeat
        '''When the file ends, play it again'''
        self.__is_playout_paused = False
        self.__is_recording_paused = False

        self.__input_filename = input_filename or ''
        self.__output_filename = output_filename or ''

        self.__file_audio_device_descriptor = None

    def __create_and_return_file_audio_device_descriptor(self):
        self.__file_audio_device_descriptor = tgcalls.FileAudioDeviceDescriptor()
        self.__file_audio_device_descriptor.getInputFilename = self.__get_input_filename_callback
        self.__file_audio_device_descriptor.getOutputFilename = self.__get_output_filename_callback
        self.__file_audio_device_descriptor.isEndlessPlayout = self.__is_endless_playout_callback
        self.__file_audio_device_descriptor.isPlayoutPaused = self.__is_playout_paused_callback
        self.__file_audio_device_descriptor.isRecordingPaused = self.__is_recording_paused_callback
        self.__file_audio_device_descriptor.playoutEndedCallback = self.__playout_ended_callback

        return self.__file_audio_device_descriptor

    def _setup_and_start_group_call(self):
        self._start_native_group_call(self.__create_and_return_file_audio_device_descriptor())

    def stop_playout(self):
        """Stop playing of file."""

        self.input_filename = ''

    def stop_output(self):
        """Stop recording to file."""

        self.output_filename = ''

    @property
    def input_filename(self):
        """Input filename (or path) to play."""

        return self.__input_filename

    @input_filename.setter
    def input_filename(self, filename):
        self.__input_filename = filename or ''
        if self.is_connected:
            self.restart_playout()

    @property
    def output_filename(self):
        """Output filename (or path) to record."""

        return self.__output_filename

    @output_filename.setter
    def output_filename(self, filename):
        self.__output_filename = filename or ''
        if self.is_connected:
            self.restart_recording()

    def pause_playout(self):
        """Pause playout (playing from file)."""
        self.__is_playout_paused = True

    def resume_playout(self):
        """Resume playout (playing from file)."""
        self.__is_playout_paused = False

    def pause_recording(self):
        """Pause recording (output to file)."""
        self.__is_recording_paused = True

    def resume_recording(self):
        """Resume recording (output to file)."""
        self.__is_recording_paused = False

    def __get_input_filename_callback(self):
        return self.__input_filename

    def __get_output_filename_callback(self):
        return self.__output_filename

    def __is_endless_playout_callback(self):
        return self.play_on_repeat

    def __is_playout_paused_callback(self):
        return self.__is_playout_paused

    def __is_recording_paused_callback(self):
        return self.__is_recording_paused

    def __playout_ended_callback(self, input_filename: str):
        self.trigger_handlers(GroupCallFileAction.PLAYOUT_ENDED, self, input_filename)



from enum import Enum
from typing import Callable, Optional

from pytgcalls.dispatcher import Action, DispatcherMixin
from pytgcalls.implementation import GroupCallRaw, GroupCallBaseAction
from pytgcalls.utils import AudioStream, VideoStream


class MediaType(Enum):
    VIDEO = 'video'
    AUDIO = 'audio'


class GroupCallAction(GroupCallBaseAction):
    VIDEO_PLAYOUT_ENDED = Action()
    '''When a video playout will be ended.'''
    AUDIO_PLAYOUT_ENDED = Action()
    '''When a audio playout will be ended.'''
    PLAYOUT_ENDED = MEDIA_PLAYOUT_ENDED = Action()
    '''When a audio or video playout will be ended.'''


class GroupCallDispatcherMixin(DispatcherMixin):
    def on_video_playout_ended(self, func: Callable) -> Callable:
        """When a video playout will be ended.

        Args:
            func (`Callable`): A functions that accept group_call and source args.

        Returns:
            `Callable`: passed to args callback function.
        """

        return self.add_handler(func, GroupCallAction.VIDEO_PLAYOUT_ENDED)

    def on_audio_playout_ended(self, func: Callable) -> Callable:
        """When a audio playout will be ended.

        Args:
            func (`Callable`): A functions that accept group_call and source args.

        Returns:
            `Callable`: passed to args callback function.
        """

        return self.add_handler(func, GroupCallAction.AUDIO_PLAYOUT_ENDED)

    def on_media_playout_ended(self, func: Callable) -> Callable:
        """When a audio or video playout will be ended.

        Args:
            func (`Callable`): A functions that accept group_call source and media type args.

        Returns:
            `Callable`: passed to args callback function.
        """

        return self.add_handler(func, GroupCallAction.MEDIA_PLAYOUT_ENDED)

    # alias
    on_playout_ended = on_media_playout_ended


class GroupCall(GroupCallRaw, GroupCallDispatcherMixin):
    def __init__(
        self,
        mtproto_bridge,
        enable_logs_to_console=False,
        path_to_log_file=None,
        outgoing_audio_bitrate_kbit=128,
    ):
        super().__init__(
            mtproto_bridge,
            self.__on_audio_played_data,
            self.__on_audio_recorded_data,
            self.__on_video_played_data,
            enable_logs_to_console,
            path_to_log_file,
            outgoing_audio_bitrate_kbit,
        )
        super(GroupCallDispatcherMixin, self).__init__(GroupCallAction)

    def __trigger_on_video_playout_ended(self, source):
        self.trigger_handlers(GroupCallAction.VIDEO_PLAYOUT_ENDED, self, source)

    def __trigger_on_audio_playout_ended(self, source):
        self.trigger_handlers(GroupCallAction.AUDIO_PLAYOUT_ENDED, self, source)

    def __trigger_on_media_playout_ended(self, source, media_type: MediaType):
        self.trigger_handlers(GroupCallAction.MEDIA_PLAYOUT_ENDED, self, source, media_type)

    def __combined_video_trigger(self, source):
        self.__trigger_on_video_playout_ended(source)
        self.__trigger_on_media_playout_ended(source, MediaType.VIDEO)

    def __combined_audio_trigger(self, source):
        self.__trigger_on_audio_playout_ended(source)
        self.__trigger_on_media_playout_ended(source, MediaType.AUDIO)

    @staticmethod
    def __on_video_played_data(self: 'GroupCallRaw') -> bytes:
        if self._video_stream:
            return self._video_stream.read()

    @staticmethod
    def __on_audio_played_data(self: 'GroupCallRaw', length: int) -> bytes:
        if self._audio_stream:
            return self._audio_stream.read(length)

    @staticmethod
    def __on_audio_recorded_data(self, data, length):
        # TODO
        pass

    async def join(self, group, join_as=None, invite_hash: Optional[str] = None, enable_action=True):
        return await self.start(group, join_as, invite_hash, enable_action)

    async def leave(self):
        return await self.stop()

    async def start_video(
        self, source: Optional[str] = None, with_audio=True, repeat=True, enable_experimental_lip_sync=False
    ):
        """Enable video playing for current group call.

        Note:
            Source is video file or image file sequence or a
            capturing device or a IP video stream for video capturing.

            To use device camera you need to pass device index as int to the `source` arg.

        Args:
            source (`str`): Path to filename or device index or URL with some protocol. For example RTCP.
            with_audio (`bool`): Get and play audio stream from video source.
            repeat (`bool`): rewind video when end of file.
            enable_experimental_lip_sync (`bool`): enable experimental lip sync feature.
        """

        if self._video_stream and self._video_stream.is_running:
            self._video_stream.stop()

        self._video_stream = VideoStream(source, repeat, self.__combined_video_trigger)
        self._configure_video_capture(self._video_stream.get_video_info())

        if with_audio:
            await self.start_audio(source, repeat, self._video_stream if enable_experimental_lip_sync else None)
        self._video_stream.start()

        self._is_video_stopped = False
        if self.is_connected:
            await self.edit_group_call(video_stopped=False)

    async def start_audio(self, source: Optional[str] = None, repeat=True, video_stream: Optional[VideoStream] = None):
        """Enable audio playing for current group call.

        Note:
            Source is audio file or direct URL to file or audio live stream.

            If the source is None then empty bytes will be sent.

        Args:
            source (`str`, optional): Path to filename or URL to audio file or URL to live stream.
            repeat (`bool`, optional): rewind audio when end of file.
            video_stream (`VideoStream`, optional): stream to sync.
        """

        if self._audio_stream and self._audio_stream.is_running:
            self._audio_stream.stop()

        if source is not None:
            self._audio_stream = AudioStream(
                source, repeat, self.__combined_audio_trigger, video_stream=video_stream
            ).start()

        if self.is_connected:
            await self.edit_group_call(muted=False)

    async def start_audio_record(self, path):
        # TODO
        pass

    async def set_video_pause(self, pause: bool, with_mtproto=True):
        self._is_video_paused = pause
        if self.is_connected and with_mtproto:
            await self.edit_group_call(video_paused=pause)
        if self._video_stream:
            self._video_stream.set_pause(pause)

    async def set_audio_pause(self, pause: bool, with_mtproto=True):
        self._is_muted = pause
        if self.is_connected and with_mtproto:
            await self.edit_group_call(muted=pause)
        if self._audio_stream:
            self._audio_stream.set_pause(pause)

    async def set_pause(self, pause: bool):
        await self.set_video_pause(pause, False)
        await self.set_audio_pause(pause, False)

        if self.is_connected:
            # optimize 2 queries into 1
            await self.edit_group_call(muted=pause, video_paused=pause)

    async def stop_audio(self, with_mtproto=True):
        if self._audio_stream and self._audio_stream.is_running:
            self._audio_stream.stop()

        if self.is_connected and with_mtproto:
            await self.edit_group_call(muted=True)

    async def stop_video(self, with_mtproto=True):
        if self._video_stream and self._video_stream.is_running:
            self._video_stream.stop()

        if self.is_connected and with_mtproto:
            await self.edit_group_call(video_stopped=True)

    async def stop_media(self):
        await self.stop_video(False)
        await self.stop_audio(False)

        if self.is_connected:
            # optimize 2 queries into 1
            await self.edit_group_call(video_stopped=True, muted=True)

    @property
    def is_video_paused(self):
        return self._video_stream and self._video_stream.is_paused

    @property
    def is_audio_paused(self):
        return self._audio_stream and self._audio_stream.is_paused

    @property
    def is_paused(self):
        return self.is_video_paused and self.is_audio_paused

    @property
    def is_video_running(self):
        return self._video_stream and self._video_stream.is_running

    @property
    def is_audio_running(self):
        return self._audio_stream and self._audio_stream.is_running

    @property
    def is_running(self):
        return self.is_video_running and self.is_audio_running



from pytgcalls.exceptions import PytgcallsError
from pytgcalls.group_call_factory import GroupCallFactory
from pytgcalls.implementation.group_call_file import GroupCallFileAction
from pytgcalls.implementation.group_call_base import GroupCallBaseAction


__all__ = [
    'GroupCallFactory',
    'GroupCallFileAction',
    'GroupCallBaseAction',
]
__version__ = '3.0.0.dev24'
__pdoc__ = {
    # files
    'utils': False,
    'dispatcher': False,
}



from pytgcalls.implementation.group_call_native import GroupCallNative

from pytgcalls.implementation.group_call_base import GroupCallBaseAction
from pytgcalls.implementation.group_call_base import GroupCallBaseDispatcherMixin
from pytgcalls.implementation.group_call_base import GroupCallBase

from pytgcalls.implementation.group_call_file import GroupCallFile
from pytgcalls.implementation.group_call_device import GroupCallDevice
from pytgcalls.implementation.group_call_raw import GroupCallRaw

__all__ = [
    'GroupCallNative',
    'GroupCallBase',
    'GroupCallBaseAction',
    'GroupCallBaseDispatcherMixin',
    'GroupCallFile',
    'GroupCallDevice',
    'GroupCallRaw',
]



import asyncio
import logging
from abc import ABC
from typing import Callable, Optional

import tgcalls

from pytgcalls.dispatcher import Action, DispatcherMixin
from pytgcalls.exceptions import GroupCallNotFoundError, NotConnectedError
from pytgcalls.implementation import GroupCallNative
from pytgcalls.mtproto.data import GroupCallDiscardedWrapper
from pytgcalls.mtproto.data.update import UpdateGroupCallParticipantsWrapper, UpdateGroupCallWrapper
from pytgcalls.mtproto.exceptions import GroupcallSsrcDuplicateMuch
from pytgcalls.utils import uint_ssrc

logger = logging.getLogger(__name__)


class GroupCallBaseAction:
    NETWORK_STATUS_CHANGED = Action()
    '''When a status of network will be changed.'''
    PARTICIPANT_LIST_UPDATED = Action()
    '''When a list of participant will be updated.'''


class GroupCallBaseDispatcherMixin(DispatcherMixin):
    def on_network_status_changed(self, func: Callable) -> Callable:
        """When a status of network will be changed.

        Args:
            func (`Callable`): A functions that accept group_call and is_connected args.

        Returns:
            `Callable`: passed to args callback function.
        """

        return self.add_handler(func, GroupCallBaseAction.NETWORK_STATUS_CHANGED)

    def on_participant_list_updated(self, func: Callable) -> Callable:
        """When a list of participant will be updated.

        Args:
            func (`Callable`): A functions that accept group_call and participants args.

        Note:
            The `participants` arg is a `list` of `GroupCallParticipantWrapper`.
            It contains only updated participants! It's not a list of all participants!

        Returns:
            `Callable`: passed to args callback function.
        """

        return self.add_handler(func, GroupCallBaseAction.PARTICIPANT_LIST_UPDATED)


class GroupCallBase(ABC, GroupCallBaseDispatcherMixin, GroupCallNative):
    SEND_ACTION_UPDATE_EACH = 0.45
    '''How often to send speaking action to chat'''

    __ASYNCIO_TIMEOUT = 10

    def __init__(
        self,
        mtproto_bridge,
        enable_logs_to_console: bool,
        path_to_log_file: str,
        outgoing_audio_bitrate_kbit: int,
    ):
        GroupCallNative.__init__(
            self,
            self.__emit_join_payload_callback,
            self.__network_state_updated_callback,
            enable_logs_to_console,
            path_to_log_file,
            outgoing_audio_bitrate_kbit,
        )
        GroupCallBaseDispatcherMixin.__init__(self, GroupCallBaseAction)

        self.mtproto = mtproto_bridge
        self.mtproto.register_group_call_native_callback(
            self._group_call_participants_update_callback, self._group_call_update_callback
        )

        self.invite_hash = None
        '''Hash from invite link to join as speaker'''

        self.enable_action = True
        '''Is enable sending of speaking action'''

        self.is_connected = False
        '''Is connected to voice chat via tgcalls'''

        self.__is_stop_requested = False
        self.__emit_join_payload_event = None

        self._is_muted = True
        self._is_video_stopped = True
        self._is_video_paused = False

        self._video_stream = None
        self._audio_stream = None

    async def _group_call_participants_update_callback(self, update: UpdateGroupCallParticipantsWrapper):
        logger.debug('Group call participants update...')
        logger.debug(update)

        self.trigger_handlers(GroupCallBaseAction.PARTICIPANT_LIST_UPDATED, self, update.participants)

        for participant in update.participants:
            ssrc = uint_ssrc(participant.source)

            # maybe (if needed) set unmute status on server side after allowing to speak by admin
            # also mb there is need a some delay after getting update cuz server sometimes cant handle editing properly
            if participant.is_self and participant.can_self_unmute:
                if not self._is_muted:
                    await self.edit_group_call(muted=False)

            if participant.peer == self.mtproto.join_as and ssrc != self.mtproto.my_ssrc:
                logger.debug(f'Not equal ssrc. Expected: {ssrc}. Actual: {self.mtproto.my_ssrc}.')
                await self.reconnect()

    async def _group_call_update_callback(self, update: UpdateGroupCallWrapper):
        logger.debug('Group call update...')
        logger.debug(update)

        if isinstance(update.call, GroupCallDiscardedWrapper):
            logger.debug('Group call discarded.')
            await self.stop()
        elif update.call.params:
            self.__set_join_response_payload(update.call.params.data)

    def __set_join_response_payload(self, payload):
        logger.debug('Set join response payload...')

        if self.__is_stop_requested:
            logger.debug('Set payload rejected by a stop request.')
            return

        self._set_connection_mode(tgcalls.GroupConnectionMode.GroupConnectionModeRtc)
        self._set_join_response_payload(payload)
        logger.debug('Join response payload was set.')

    def __emit_join_payload_callback(self, payload):
        logger.debug('Emit join payload callback...')

        if self.__is_stop_requested:
            logger.debug('Join group call rejected by a stop request.')
            return

        if self.mtproto.group_call is None:
            logger.debug('Group Call is None.')
            return

        async def _():
            try:

                def pre_update_processing():
                    logger.debug(f'Set my ssrc to {payload.audioSsrc}.')
                    self.mtproto.set_my_ssrc(payload.audioSsrc)

                await self.mtproto.join_group_call(
                    self.invite_hash,
                    payload.json,
                    muted=True,
                    video_stopped=self._is_video_stopped,
                    pre_update_processing=pre_update_processing,
                )

                if self.__emit_join_payload_event:
                    self.__emit_join_payload_event.set()

                logger.debug(
                    f'Successfully connected to VC with '
                    f'ssrc={self.mtproto.my_ssrc} '
                    f'as {type(self.mtproto.join_as).__name__}.'
                )
            except GroupcallSsrcDuplicateMuch:
                logger.debug('Duplicate SSRC.')
                await self.reconnect()

        self.get_event_loop().create_task(_())

    def __network_state_updated_callback(self, state: bool):
        logger.debug('Network state updated...')

        if self.is_connected == state:
            logger.debug('Network state is same. Do nothing.')
            return

        self.is_connected = state
        if self.is_connected:
            self.get_event_loop().create_task(self.set_is_mute(False))
            if self.enable_action:
                self.__start_status_worker()

        self.trigger_handlers(GroupCallBaseAction.NETWORK_STATUS_CHANGED, self, state)

        logger.debug(f'New network state is {self.is_connected}.')

    def __start_status_worker(self):
        async def worker():
            logger.debug('Start status (call action) worker...')
            while self.is_connected:
                await self.mtproto.send_speaking_group_call_action()
                await asyncio.sleep(self.SEND_ACTION_UPDATE_EACH)

        self.get_event_loop().create_task(worker())

    async def start(self, group, join_as=None, invite_hash: Optional[str] = None, enable_action=True):
        """Start voice chat (join and play/record from initial values).

        Note:
            Disconnect from current voice chat and connect to the new one.
            Multiple instances of `GroupCallBase` must be created for multiple voice chats at the same time.
            Join as by default is personal account.

        Args:
            group (`InputPeerChannel` | `InputPeerChat` | `str` | `int`): Chat ID in any form.
            join_as (`InputPeer` | `str` | `int`, optional): How to present yourself in participants list.
            invite_hash (`str`, optional): Hash from speaker invite link.
            enable_action (`bool`, optional): Is enables sending of speaking action.
        """
        self.__is_stop_requested = False
        self.enable_action = enable_action

        group_call = await self.mtproto.get_and_set_group_call(group)
        if group_call is None:
            raise GroupCallNotFoundError('Chat without a voice chat')

        # mb move in other place. save plain join_as arg and use it in JoinGroupCall
        # but for now it works  as optimization of requests
        # we resolve join_as only when try to connect
        # it doesnt call resolve on reconnect
        await self.mtproto.resolve_and_set_join_as(join_as)

        self.invite_hash = invite_hash

        self.mtproto.re_register_update_handlers()

        # when trying to connect to another chat or with another join_as
        if self.is_group_call_native_created():
            await self.reconnect()
        # the first start
        else:
            self._setup_and_start_group_call()

    async def stop(self):
        """Properly stop tgcalls, remove MTProto handler, leave from server side."""
        if not self.is_group_call_native_created():
            logger.debug('Group call is not started, so there\'s nothing to stop.')
            return

        self.__is_stop_requested = True
        logger.debug('Stop requested.')

        self.mtproto.unregister_update_handlers()
        # to bypass recreating of outgoing audio channel
        self._set_is_mute(True)
        self._set_connection_mode(tgcalls.GroupConnectionMode.GroupConnectionModeNone)

        on_disconnect_event = asyncio.Event()

        async def post_disconnect():
            await self.leave_current_group_call()
            self.mtproto.reset()
            self.__is_stop_requested = False

        async def on_disconnect(obj, is_connected):
            if is_connected:
                return

            obj._stop_audio_device_module()

            # need for normal waiting of stopping audio devices
            # destroying of webrtc client during .stop not needed yet
            # because we a working in the same native instance
            # and can reuse tis client for another connections.
            # In any case now its possible to reset group call ptr
            # self.__native_instance.stopGroupCall()

            await post_disconnect()

            obj.remove_handler(on_disconnect, GroupCallBaseAction.NETWORK_STATUS_CHANGED)
            on_disconnect_event.set()

        if self.is_connected:
            self.add_handler(on_disconnect, GroupCallBaseAction.NETWORK_STATUS_CHANGED)
            await asyncio.wait_for(on_disconnect_event.wait(), timeout=self.__ASYNCIO_TIMEOUT)
        else:
            await post_disconnect()

        if self._video_stream and self._video_stream.is_running:
            self._video_stream.stop()
        if self._audio_stream and self._audio_stream.is_running:
            self._audio_stream.stop()

        logger.debug('GroupCallBase stopped properly.')

    async def reconnect(self):
        """Connect to voice chat using the same native instance."""
        logger.debug('Reconnecting...')
        if not self.mtproto.group_call:
            raise NotConnectedError("You don't connected to voice chat.")

        self._set_connection_mode(tgcalls.GroupConnectionMode.GroupConnectionModeNone)
        self._emit_join_payload(self.__emit_join_payload_callback)

        # during the .stop we stop audio device module. Need to restart
        self.restart_recording()
        self.restart_playout()

        # cuz native instance doesnt block python
        self.__emit_join_payload_event = asyncio.Event()
        await asyncio.wait_for(self.__emit_join_payload_event.wait(), timeout=self.__ASYNCIO_TIMEOUT)

    async def leave_current_group_call(self):
        """Leave group call from server side (MTProto part)."""
        logger.debug('Try to leave the current group call...')
        try:
            await self.mtproto.leave_current_group_call()
        except Exception as e:
            logger.warning("Couldn't leave the group call. But no worries, you'll get removed from it in seconds.")
            logger.debug(e)
        else:
            logger.debug('Completely left the current group call.')

    async def edit_group_call(self, volume: int = None, muted=None, video_stopped=None, video_paused=None):
        """Edit own settings of group call.

        Note:
            There is bug where you can try to pass `volume=100`.

        Args:
            volume (`int`): Volume.
            muted (`bool`): Is muted.
            video_stopped (`bool`): Is video stopped.
            video_paused (`bool`): Is video paused.
        """

        if not muted:
            muted = self._is_muted

        if video_stopped is None:
            video_stopped = self._is_video_stopped
        else:
            self._is_video_stopped = video_stopped

        if video_paused is None:
            video_paused = self._is_video_paused
        else:
            self._is_video_paused = video_paused

        await self.edit_group_call_member(self.mtproto.join_as, volume, muted, video_stopped, video_paused)

    async def edit_group_call_member(self, peer, volume: int = None, muted=None, video_stopped=None, video_paused=None):
        """Edit setting of user in voice chat (required voice chat management permission).

        Note:
            There is bug where you can try to pass `volume=100`.

        Args:
            peer (`InputPeer`): Participant of voice chat.
            volume (`int`): Volume.
            muted (`bool`): Is muted.
            video_stopped (`bool`): Is video stopped.
            video_paused (`bool`): Is video paused.
        """

        volume = max(1, volume * 100) if volume is not None else None
        await self.mtproto.edit_group_call_member(peer, volume, muted, video_stopped, video_paused)

    async def set_is_mute(self, is_muted: bool):
        """Set is mute.

        Args:
            is_muted (`bool`): Is muted.
        """

        self._is_muted = is_muted
        self._set_is_mute(is_muted)

        logger.debug(f'Set is muted on server side. New value: {is_muted}.')
        await self.edit_group_call(muted=is_muted)

    async def set_my_volume(self, volume):
        """Set volume for current client.

        Note:
            Volume value only can be in 1-200 range. There is auto normalization.

        Args:
            volume (`int` | `str` | `float`): Volume.
        """
        # Required "Manage Voice Chats" admin permission

        volume = max(1, min(int(volume), 200))
        logger.debug(f'Set volume to: {volume}.')

        await self.edit_group_call(volume)
        self._set_volume(uint_ssrc(self.mtproto.my_ssrc), volume / 100)

    # shortcuts for easy access in callbacks of events

    @property
    def client(self):
        return self.mtproto.client

    @property
    def full_chat(self):
        return self.mtproto.full_chat

    @property
    def chat_peer(self):
        return self.mtproto.chat_peer

    @property
    def group_call(self):
        return self.mtproto.group_call

    @property
    def my_ssrc(self):
        return self.mtproto.my_ssrc

    @property
    def my_peer(self):
        return self.mtproto.my_peer

    @property
    def join_as(self):
        return self.mtproto.join_as



from typing import Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from . import GroupCallNative

from .dispatcher import Dispatcher


class DispatcherMixin:
    def __init__(self, actions):
        self._dispatcher = Dispatcher(actions)

    def add_handler(self, callback: Callable, action: str) -> Callable:
        """Register new handler.

        Args:
            callback (`Callable`): Callback function.
            action (`str`): Action.

        Returns:
            `Callable`: original callback.
        """

        return self._dispatcher.add_handler(callback, action)

    def remove_handler(self, callback: Callable, action: str) -> bool:
        """Unregister the handler.

        Args:
            callback (`Callable`): Callback function.
            action (`str`): Action.

        Returns:
            `bool`: Return `True` if success.
        """

        return self._dispatcher.remove_handler(callback, action)

    def trigger_handlers(self, action: str, instance: 'GroupCallNative', *args, **kwargs):
        """Unregister the handler.

        Args:
            action (`str`): Action.
            instance (`GroupCallNative`): Instance of GroupCallBase.
            *args (`list`, optional): Arbitrary callback arguments.
            **kwargs (`dict`, optional): Arbitrary callback arguments.
        """

        self._dispatcher.trigger_handlers(action, instance, *args, **kwargs)



from datetime import datetime
from typing import Optional, Union

from pytgcalls.mtproto.data import WrapperBase


class GroupCallParticipantWrapper(WrapperBase):
    """Group Call Participant wrapper for any MTProto client.

    Note:
        `peer` will be `raw.base.Peer` when you are using Pyrogram bridge.

        `peer` will be `TypePeer` when you are using Telethon bridge.

        `date` and `active_date`will be `int` when you are using Pyrogram bridge.

        `date` and `active_date`will be `datetime` when you are using Telethon bridge.

        `video_joined` always will be `None` for Pyrogram until it will update.
    """

    def __init__(
        self,
        peer: Union['raw.base.Peer', 'TypePeer'],
        date: Optional[Union[int, datetime]],
        source: int,
        muted: Optional[bool] = None,
        left: Optional[bool] = None,
        can_self_unmute: Optional[bool] = None,
        just_joined: Optional[bool] = None,
        versioned: Optional[bool] = None,
        min: Optional[bool] = None,
        muted_by_you: Optional[bool] = None,
        volume_by_admin: Optional[bool] = None,
        is_self: Optional[bool] = None,
        video_joined: Optional[bool] = None,
        active_date: Optional[Union[int, datetime]] = None,
        volume: Optional[int] = None,
        about: Optional[str] = None,
        raise_hand_rating: Optional[int] = None,
        **kwargs,  # to bypass new attr from layers
    ):
        self.peer = peer
        '''Peer (as joined) of the group call participant'''
        self.date = date
        self.source = source
        '''User's audio channel synchronization source identifier'''
        self.muted = muted
        '''True, if the participant is muted for all users'''
        self.left = left
        '''True, if the participant left the group call'''
        self.can_self_unmute = can_self_unmute
        '''True, if the participant is muted for all users, but can unmute themselves'''
        self.just_joined = just_joined
        self.versioned = versioned
        self.min = min
        self.muted_by_you = muted_by_you
        '''True, if the current user can mute the participant only for self'''
        self.volume_by_admin = volume_by_admin
        self.is_self = is_self
        '''True, if the participant is the current user'''
        self.video_joined = video_joined
        self.active_date = active_date
        self.volume = volume
        '''Participant's volume level'''
        self.about = about
        '''The participant user's bio or the participant chat's description'''
        self.raise_hand_rating = raise_hand_rating

    @classmethod
    def create(cls, participant):
        if hasattr(participant, '__slots__'):  # pyrogram
            attrs = participant.__slots__
            args_for_init = {}
            for attr in attrs:
                args_for_init[attr] = getattr(participant, attr, None)

            return cls(**args_for_init)
        else:  # telethon
            return cls(**vars(participant))



from typing import Callable

from pyrogram.errors import (
    BadRequest as PyrogramBadRequest,
    GroupcallSsrcDuplicateMuch as PyrogramGroupcallSsrcDuplicateMuch,
)
from pyrogram.handlers import RawUpdateHandler
from pyrogram.raw import functions, types
from pyrogram.raw.types import (
    GroupCallDiscarded as PyrogramGroupCallDiscarded,
    InputPeerChannel,
    InputPeerChat,
    UpdateGroupCallConnection,
)
from pyrogram.utils import get_peer_id

from pytgcalls import PytgcallsError
from pytgcalls.mtproto.data import GroupCallDiscardedWrapper, GroupCallWrapper, GroupCallParticipantWrapper
from pytgcalls.mtproto.data.update import UpdateGroupCallWrapper, UpdateGroupCallParticipantsWrapper
from pytgcalls.mtproto.exceptions import BadRequest, GroupcallSsrcDuplicateMuch
from pytgcalls.utils import int_ssrc

from pyrogram import Client, ContinuePropagation

from pytgcalls.mtproto import MTProtoBridgeBase


class PyrogramBridge(MTProtoBridgeBase):
    def __init__(self, client: Client):
        super().__init__(client)

        # compatibility with pyro > 2.0
        if getattr(client, 'send', None) is None:
            setattr(client, 'send', getattr(client, 'invoke'))

        self._update_to_handler = {
            types.UpdateGroupCallParticipants: self._process_group_call_participants_update,
            types.UpdateGroupCall: self._process_group_call_update,
        }

        self._handler_group = None
        self._update_handler = RawUpdateHandler(self._process_update)

    async def _process_update(self, _, update, users, chats):
        if type(update) not in self._update_to_handler.keys():
            raise ContinuePropagation

        if not self.group_call or not update.call or update.call.id != self.group_call.id:
            raise ContinuePropagation
        self.group_call = update.call

        await self._update_to_handler[type(update)](update)

    async def _process_group_call_participants_update(self, update):
        participants = [GroupCallParticipantWrapper.create(p) for p in update.participants]
        wrapped_update = UpdateGroupCallParticipantsWrapper(participants)

        await self.group_call_participants_update_callback(wrapped_update)

    async def _process_group_call_update(self, update):
        if not isinstance(update.call, PyrogramGroupCallDiscarded):
            return

        call = GroupCallDiscardedWrapper()  # no info needed
        wrapped_update = UpdateGroupCallWrapper(update.chat_id, call)

        await self.group_call_update_callback(wrapped_update)

    async def _process_group_call_connection(self, update):
        # TODO update to new layer when pyrogram will release new stable version on pypi
        call = GroupCallWrapper('placeholder', update.params)
        wrapped_update = UpdateGroupCallWrapper('placeholder', call)

        await self.group_call_update_callback(wrapped_update)

    async def check_group_call(self) -> bool:
        if not self.group_call or not self.my_ssrc:
            return False

        try:
            in_group_call = await (
                self.client.send(functions.phone.CheckGroupCall(call=self.group_call, source=int_ssrc(self.my_ssrc)))
            )
        except PyrogramBadRequest as e:
            # compatibility with pyro > 2.0
            if getattr(e, 'x', None) is None:
                setattr(e, 'x', getattr(e, 'value'))

            if e.x != '[400 GROUPCALL_JOIN_MISSING]':
                raise BadRequest(e.x)

            in_group_call = False

        return in_group_call

    async def leave_current_group_call(self):
        if not self.full_chat or not self.full_chat.call or not self.my_ssrc:
            return

        response = await self.client.send(
            functions.phone.LeaveGroupCall(call=self.full_chat.call, source=int_ssrc(self.my_ssrc))
        )
        await self.handle_updates(response)

    async def edit_group_call_member(
        self, peer, volume: int = None, muted=False, video_stopped=True, video_paused=False
    ):
        response = await self.client.send(
            functions.phone.EditGroupCallParticipant(
                call=self.full_chat.call,
                participant=peer,
                muted=muted,
                volume=volume,
                video_stopped=video_stopped,
                video_paused=video_paused,
            )
        )
        await self.handle_updates(response)

    async def get_and_set_self_peer(self):
        self.my_peer = await self.client.resolve_peer(await self.client.storage.user_id())

        return self.my_peer

    async def get_and_set_group_call(self, group):
        """Get group call input of chat.

        Args:
            group (`InputPeerChannel` | `InputPeerChat` | `str` | `int`): Chat ID in any form.

        Returns:
            `InputGroupCall`.
        """

        self.chat_peer = group
        if type(group) not in [InputPeerChannel, InputPeerChat]:
            self.chat_peer = await self.client.resolve_peer(group)

        if isinstance(self.chat_peer, InputPeerChannel):
            self.full_chat = (
                await (self.client.send(functions.channels.GetFullChannel(channel=self.chat_peer)))
            ).full_chat
        elif isinstance(self.chat_peer, InputPeerChat):
            self.full_chat = (
                await (self.client.send(functions.messages.GetFullChat(chat_id=self.chat_peer.chat_id)))
            ).full_chat

        if self.full_chat is None:
            raise PytgcallsError(f'Can\'t get full chat by {group}')

        self.group_call = self.full_chat.call

        return self.group_call

    def unregister_update_handlers(self):
        if self._handler_group:
            self.client.remove_handler(self._update_handler, self._handler_group)
            self._handler_group = None

    def register_update_handlers(self):
        if self.group_call.id > 0:
            self._handler_group = -self.group_call.id
        self._handler_group = self.group_call.id

        self.client.add_handler(self._update_handler, self._handler_group)

    async def resolve_and_set_join_as(self, join_as):
        my_peer = await self.get_and_set_self_peer()

        if join_as is None:
            self.join_as = self.full_chat.groupcall_default_join_as
            if self.join_as:
                # convert Peer to InputPeer
                self.join_as = await self.client.resolve_peer(get_peer_id(self.join_as))
            else:
                self.join_as = my_peer
        elif isinstance(join_as, str) or isinstance(join_as, int):
            self.join_as = await self.client.resolve_peer(join_as)
        else:
            self.join_as = join_as

    async def send_speaking_group_call_action(self):
        await self.client.send(
            functions.messages.SetTyping(peer=self.chat_peer, action=types.SpeakingInGroupCallAction())
        )

    async def join_group_call(
        self, invite_hash: str, params: str, muted: bool, video_stopped: bool, pre_update_processing: Callable
    ):
        try:
            response = await self.client.send(
                functions.phone.JoinGroupCall(
                    call=self.group_call,
                    join_as=self.join_as,
                    invite_hash=invite_hash,
                    params=types.DataJSON(data=params),
                    muted=muted,
                    video_stopped=video_stopped,
                )
            )

            pre_update_processing()

            # it is here cuz we need to associate params for connection with group call
            for update in response.updates:
                if isinstance(update, UpdateGroupCallConnection):
                    await self._process_group_call_connection(update)

            await self.handle_updates(response)
        except PyrogramGroupcallSsrcDuplicateMuch as e:
            # compatibility with pyro > 2.0
            if getattr(e, 'x', None) is None:
                setattr(e, 'x', getattr(e, 'value'))

            raise GroupcallSsrcDuplicateMuch(e.x)

    async def handle_updates(self, updates):
        await self.client.handle_updates(updates)



import importlib
from importlib.util import find_spec
from typing import Callable, Optional, Union

from pytgcalls.exceptions import PytgcallsBaseException, PytgcallsError
from pytgcalls.group_call_type import GroupCallType
from pytgcalls.implementation.group_call import GroupCall
from pytgcalls.mtproto_client_type import MTProtoClientType
from pytgcalls.implementation.group_call_file import GroupCallFile
from pytgcalls.implementation.group_call_device import GroupCallDevice
from pytgcalls.implementation.group_call_raw import GroupCallRaw


def hot_load_mtproto_lib_or_exception(module):
    if find_spec(module):
        importlib.import_module(module)
    else:
        raise PytgcallsBaseException(
            f'To use this MTProto client type you need to install {module.capitalize()}. '
            f'Run this command: pip3 install -U pytgcalls[{module}]'
        )


class GroupCallFactory:
    MTPROTO_CLIENT_TYPE = MTProtoClientType
    GROUP_CALL_TYPE = GroupCallType

    GROUP_CALL_CLASS_TO_TYPE = {
        GROUP_CALL_TYPE.FILE: GroupCallFile,
        GROUP_CALL_TYPE.DEVICE: GroupCallDevice,
        GROUP_CALL_TYPE.RAW: GroupCallRaw,
    }

    def __init__(
        self,
        client,
        mtproto_backend=MTProtoClientType.PYROGRAM,
        enable_logs_to_console=False,
        path_to_log_file=None,
        outgoing_audio_bitrate_kbit=128,
    ):
        self.client = client

        if mtproto_backend is MTProtoClientType.PYROGRAM:
            hot_load_mtproto_lib_or_exception(MTProtoClientType.PYROGRAM.value)
            from pytgcalls.mtproto.pyrogram_bridge import PyrogramBridge

            self.__mtproto_bride_class = PyrogramBridge
        elif mtproto_backend is MTProtoClientType.TELETHON:
            hot_load_mtproto_lib_or_exception(MTProtoClientType.TELETHON.value)
            from pytgcalls.mtproto.telethon_bridge import TelethonBridge

            self.__mtproto_bride_class = TelethonBridge
        else:
            raise PytgcallsError('Unknown MTProto client type')

        self.enable_logs_to_console = enable_logs_to_console
        self.path_to_log_file = path_to_log_file
        self.outgoing_audio_bitrate_kbit = outgoing_audio_bitrate_kbit

    def get_mtproto_bridge(self):
        return self.__mtproto_bride_class(self.client)

    def get(self, group_call_type: GroupCallType, **kwargs) -> Union[GroupCallFile, GroupCallDevice, GroupCallRaw]:
        return GroupCallFactory.GROUP_CALL_CLASS_TO_TYPE[group_call_type](
            mtproto_bridge=self.get_mtproto_bridge(),
            enable_logs_to_console=self.enable_logs_to_console,
            path_to_log_file=self.path_to_log_file,
            outgoing_audio_bitrate_kbit=self.outgoing_audio_bitrate_kbit,
            **kwargs,
        )

    def get_group_call(self) -> GroupCall:
        return GroupCall(
            self.get_mtproto_bridge(),
            self.enable_logs_to_console,
            self.path_to_log_file,
            self.outgoing_audio_bitrate_kbit,
        )

    def get_file_group_call(
        self, input_filename: Optional[str] = None, output_filename: Optional[str] = None, play_on_repeat=True
    ) -> GroupCallFile:
        return GroupCallFile(
            self.get_mtproto_bridge(),
            input_filename,
            output_filename,
            play_on_repeat,
            self.enable_logs_to_console,
            self.path_to_log_file,
            self.outgoing_audio_bitrate_kbit,
        )

    def get_device_group_call(
        self, audio_input_device: Optional[str] = None, audio_output_device: Optional[str] = None
    ) -> GroupCallDevice:
        return GroupCallDevice(
            self.get_mtproto_bridge(),
            audio_input_device,
            audio_output_device,
            self.enable_logs_to_console,
            self.path_to_log_file,
            self.outgoing_audio_bitrate_kbit,
        )

    def get_raw_group_call(
        self,
        on_played_data: Callable[['GroupCallRaw', int], bytes] = None,
        on_recorded_data: Callable[['GroupCallRaw', bytes, int], None] = None,
        on_video_played_data: Callable[['GroupCallRaw'], bytes] = None,
    ) -> GroupCallRaw:
        return GroupCallRaw(
            self.get_mtproto_bridge(),
            on_played_data,
            on_recorded_data,
            on_video_played_data,
            self.enable_logs_to_console,
            self.path_to_log_file,
            self.outgoing_audio_bitrate_kbit,
        )



from typing import TYPE_CHECKING, Union

from pytgcalls.mtproto.data import WrapperBase

if TYPE_CHECKING:
    from pytgcalls.mtproto.data import GroupCallWrapper, GroupCallDiscardedWrapper


class UpdateGroupCallWrapper(WrapperBase):
    def __init__(self, chat_id: int, call: Union['GroupCallWrapper', 'GroupCallDiscardedWrapper']):
        self.chat_id = chat_id
        self.call = call



from abc import ABC
from asyncio import AbstractEventLoop
from typing import Callable


class MTProtoBridgeBase(ABC):
    def __init__(self, client):
        self.client = client
        '''Any MTProto client. Pyrogram/Telethon and so on'''

        self.group_call_participants_update_callback = None
        '''Native handler of wrapped group call participants update'''
        self.group_call_update_callback = None
        '''Native handler of wrapped group call update'''

        self.full_chat = None
        '''Full chat information'''
        self.chat_peer = None
        '''Chat peer where bot is now'''

        self.group_call = None
        '''Instance of MTProto's group call'''

        self.my_ssrc = None
        '''Client SSRC (Synchronization Source)'''
        self.my_peer = None
        '''Client user peer'''

        self.join_as = None
        '''How to present yourself in participants list'''

    def reset(self):
        self.group_call = self.full_chat = self.chat_peer = self.join_as = self.my_peer = self.my_ssrc = None

    def register_group_call_native_callback(
        self, group_call_participants_update_callback: Callable, group_call_update_callback: Callable
    ):
        self.group_call_participants_update_callback = group_call_participants_update_callback
        self.group_call_update_callback = group_call_update_callback

    def re_register_update_handlers(self):
        """Delete and add pytgcalls handler in MTProto client."""
        self.unregister_update_handlers()
        self.register_update_handlers()

    async def check_group_call(self) -> bool:
        """Check if client is in a voice chat.

        Returns:
            `bool`: Is in voice chat by opinion of Telegram server.

        raise wrapped BadRequest if you got [400 GROUPCALL_JOIN_MISSING] response!
        """

        raise NotImplementedError

    async def leave_current_group_call(self):
        """
        call phone.LeaveGroupCall and handle returned updates
        """
        raise NotImplementedError

    async def edit_group_call_member(self, peer, volume: int = None, muted=False):
        """
        call phone.EditGroupCallParticipant
        """
        raise NotImplementedError

    async def get_and_set_self_peer(self):
        """
        resolve self peer and set to obj field
        """
        raise NotImplementedError

    async def get_and_set_group_call(self, group):
        """
        there is group arg can be peer, int, string with username and so on
        need to support all of them

        in this method be set chat_peer, full_chat and group_call class fields

        I think group_call don't need to wrap cuz it will be used in phone.JoinGroupCall
        """
        raise NotImplementedError

    def unregister_update_handlers(self):
        """
        delete all registered handlers from MTProto client
        """
        raise NotImplementedError

    def register_update_handlers(self):
        """
        register handlers
        """
        raise NotImplementedError

    async def resolve_and_set_join_as(self, join_as):
        """
        join_as arg can be str on peer. if it str we need to resolve peer
        save join_as to class field
        """
        raise NotImplementedError

    async def send_speaking_group_call_action(self):
        """
        call messages.SetTyping with SpeakingInGroupCallAction by chat_peer
        """
        raise NotImplementedError

    async def join_group_call(
        self, invite_hash: str, params: dict, muted: bool, video_stopped: bool, pre_update_processing: Callable
    ):
        """
        call phone.JoinGroupCall with group_call, join_as, invite hash, muted and params

        handle updates from response!

        reraise wrapped GroupcallSsrcDuplicateMuch!
        """
        raise NotImplementedError

    def set_my_ssrc(self, ssrc):
        self.my_ssrc = ssrc

    def handle_updates(self, updates):
        raise NotImplementedError



from typing import Optional

from pytgcalls.implementation import GroupCallBase


class GroupCallDevice(GroupCallBase):
    def __init__(
        self,
        mtproto_bridge,
        audio_input_device: Optional[str] = None,
        audio_output_device: Optional[str] = None,
        enable_logs_to_console=False,
        path_to_log_file=None,
        outgoing_audio_bitrate_kbit=128,
    ):
        super().__init__(mtproto_bridge, enable_logs_to_console, path_to_log_file, outgoing_audio_bitrate_kbit)

        self.__is_playout_paused = False
        self.__is_recording_paused = False

        self.__audio_input_device = audio_input_device or ''
        self.__audio_output_device = audio_output_device or ''

    def _setup_and_start_group_call(self):
        self._start_native_group_call(self.__audio_input_device, self.__audio_output_device)

    @property
    def audio_input_device(self):
        """Get audio input device name or GUID

        Note:
            To get system recording device list you can use `get_recording_devices()` method.
        """

        return self.__audio_input_device

    @audio_input_device.setter
    def audio_input_device(self, name=None):
        self.set_audio_input_device(name)

    @property
    def audio_output_device(self):
        """Get audio output device name or GUID

        Note:
            To get system playout device list you can use `get_playout_devices()` method.
        """

        return self.__audio_output_device

    @audio_output_device.setter
    def audio_output_device(self, name=None):
        self.set_audio_output_device(name)



import asyncio
import logging
from typing import Callable, Optional, List

import tgcalls
from pytgcalls.exceptions import CallBeforeStartError

logger = logging.getLogger(__name__)


def if_native_instance_created(func):
    def wrapper(self, *args, **kwargs):
        if self.is_group_call_native_created():
            return func(self, *args, **kwargs)
        else:
            raise CallBeforeStartError("You can't use this method before calling .start()")

    return wrapper


class GroupCallNative:
    def __init__(
        self,
        emit_join_payload_callback,
        network_state_updated_callback,
        enable_logs_to_console: bool,
        path_to_log_file: str,
        outgoing_audio_bitrate_kbit: int,
    ):
        """Create NativeInstance of tgcalls C++ part.

        Args:
            enable_logs_to_console (`bool`): Is enable logs to stderr from tgcalls.
            path_to_log_file (`str`, optional): Path to log file for logs of tgcalls.
        """

        # bypass None value
        if not path_to_log_file:
            path_to_log_file = ''

        logger.debug('Create a new native instance...')
        self.__native_instance = tgcalls.NativeInstance(enable_logs_to_console, path_to_log_file)

        self.__native_instance.setupGroupCall(
            emit_join_payload_callback,
            network_state_updated_callback,
            outgoing_audio_bitrate_kbit,
        )

        logger.debug('Native instance created.')

        self.__loop = asyncio.get_event_loop()

    def get_event_loop(self):
        return self.__loop

    def is_group_call_native_created(self):
        return self.__native_instance.isGroupCallNativeCreated()

    def _setup_and_start_group_call(self):
        raise NotImplementedError()

    @if_native_instance_created
    def _set_connection_mode(self, mode: tgcalls.GroupConnectionMode, keep_broadcast_if_was_enabled=False):
        logger.debug(f'Set native connection mode {mode}.')
        self.__native_instance.setConnectionMode(mode, keep_broadcast_if_was_enabled)

    @if_native_instance_created
    def _emit_join_payload(self, callback):
        logger.debug(f'Trigger native emit join payload.')
        self.__native_instance.emitJoinPayload(callback)

    def _start_native_group_call(self, *args):
        logger.debug('Start native group call...')
        self.__native_instance.startGroupCall(*args)

    @if_native_instance_created
    def _emit_join_payload(self, callback):
        logger.debug('Emit native join payload.')
        self.__native_instance.emitJoinPayload(callback)

    @if_native_instance_created
    def _set_join_response_payload(self, payload):
        logger.debug('Set native join response payload...')
        self.__native_instance.setJoinResponsePayload(payload)

    @if_native_instance_created
    def _set_is_mute(self, is_muted: bool):
        logger.debug(f'Set is muted on native instance side. New value: {is_muted}.')
        self.__native_instance.setIsMuted(is_muted)

    @if_native_instance_created
    def _set_volume(self, ssrc, volume):
        logger.debug(f'Set native volume for {ssrc} to {volume}.')
        self.__native_instance.setVolume(ssrc, volume)

    @if_native_instance_created
    def _stop_audio_device_module(self):
        logger.debug(f'Stop audio device module.')
        self.__native_instance.stopAudioDeviceModule()

    @if_native_instance_created
    def _start_audio_device_module(self):
        logger.debug(f'Start audio device module.')
        self.__native_instance.startAudioDeviceModule()

    @if_native_instance_created
    def _set_video_capture(self, source_path: Callable, width: int, height: int, fps: int):
        logger.debug('Set video capture.')
        self.__native_instance.setVideoCapture(source_path, fps, width, height)

    @if_native_instance_created
    def get_playout_devices(self) -> List['tgcalls.AudioDevice']:
        """Get available playout audio devices in the system.

        Note:
            `tgcalls.AudioDevice` have 2 attributes: name, guid.
        """

        logger.debug('Get native playout devices.')
        return self.__native_instance.getPlayoutDevices()

    @if_native_instance_created
    def get_recording_devices(self) -> List['tgcalls.AudioDevice']:
        """Get available recording audio devices in the system.

        Note:
            `tgcalls.AudioDevice` have 2 attributes: name, guid.
        """

        logger.debug('Get native recording devices.')
        return self.__native_instance.getRecordingDevices()

    @if_native_instance_created
    def set_audio_input_device(self, name: Optional[str] = None):
        """Set audio input device.

        Note:
            If `name` is `None`, will use default system device.
            And this is works only at first device initialization time!

        Args:
            name (`str`): Name or GUID of device.
        """

        logger.debug(f'Set native audio input device to {name}.')
        self.__native_instance.setAudioInputDevice(name or '')

    @if_native_instance_created
    def set_audio_output_device(self, name: Optional[str] = None):
        """Set audio output device.

        Note:
            If `name` is `None`, will use default system device.
            And this is works only at first device initialization time!

        Args:
            name (`str`): Name or GUID of device.
        """

        logger.debug(f'Set native audio output device to {name}.')
        self.__native_instance.setAudioOutputDevice(name or '')

    @if_native_instance_created
    def restart_playout(self):
        """Start play current input file from start or just reload file audio device.

        Note:
            Device restart needed to apply new filename in tgcalls.
        """

        logger.debug(f'Restart native audio input device.')
        self.__native_instance.restartAudioInputDevice()

    @if_native_instance_created
    def restart_recording(self):
        """Start recording to output file from begin or just restart recording device.

        Note:
            Device restart needed to apply new filename in tgcalls.
        """

        logger.debug(f'Restart native audio output device.')
        self.__native_instance.restartAudioOutputDevice()



import os
from logging import getLogger
from queue import Empty, Queue
from threading import Lock, Thread
from typing import List, Optional

import cv2
import av


logger = getLogger(__name__)

uint_ssrc = lambda ssrc: ssrc if ssrc >= 0 else ssrc + 2 ** 32
int_ssrc = lambda ssrc: ssrc if ssrc < 2 ** 31 else ssrc - 2 ** 32

# increasing this value will increase memory usage
VIDEO_QUEUE_SIZE = 1  # 1 frame depends on FPS
AUDIO_QUEUE_SIZE = 1  # 1 frame is 10 ms

DEFAULT_COLOR_DEPTH = 4
DEFAULT_REQUESTED_AUDIO_BYTES_LENGTH = 960
DEFAULT_AUDIO_SAMPLE_RATE = 48000
REVERSED_AUDIO_SAMPLE_RATE = {
    48000: 44100,
    44100: 48000,
}

MILLIS_IN_SEC = 1000


class VideoInfo:
    def __init__(self, width: int, height: int, fps: int):
        self.width = width
        self.height = height
        self.fps = fps

    @classmethod
    def default(cls):
        return cls(1820, 720, 30)


class QueueStream:
    def __init__(self, on_end_callback, queue_size):
        self.queue_size = queue_size
        self.__queue = self.create_queue()

        self.thread = Thread(target=self._update, args=())
        self.thread.daemon = True

        self.is_running = False
        self.is_paused = False

        self._on_end_callback = on_end_callback

    def set_pause(self, pause: bool):
        self.is_paused = pause

    def create_queue(self, queue_size=None):
        if not queue_size:
            queue_size = self.queue_size

        return Queue(maxsize=queue_size)

    def start(self):
        self.is_running = True
        self.thread.start()
        return self

    def stop(self):
        self.is_running = False
        self.__queue = self.create_queue()

    def read(self):
        return self.__queue.get_nowait()

    def put(self, item, block=True):
        self.__queue.put(item, block)

    def _update(self):
        raise NotImplementedError


class VideoStream(QueueStream):
    def __init__(self, source, repeat, on_end_callback, queue_size=VIDEO_QUEUE_SIZE):
        super().__init__(on_end_callback, queue_size)

        self.source = source
        self.video_capture = None
        if source:
            self.video_capture = cv2.VideoCapture(source)
            self.video_capture.set(cv2.CAP_PROP_BUFFERSIZE, 1)

        self.repeat = repeat

        self.__pts = 0
        self.__last_frame = None
        self.__lock = Lock()

    def start(self):
        return super().start()

    def get_video_info(self) -> VideoInfo:
        if self.video_capture and self.video_capture.isOpened():
            return VideoInfo(
                round(self.video_capture.get(cv2.CAP_PROP_FRAME_WIDTH)),
                round(self.video_capture.get(cv2.CAP_PROP_FRAME_HEIGHT)),
                self.video_capture.get(cv2.CAP_PROP_FPS),
            )

        return VideoInfo.default()

    def read(self):
        try:
            if self.is_paused:
                raise Empty

            return super().read()
        except Empty:
            if self.__last_frame:
                return self.__last_frame

    def get_pts(self):
        return self.__pts

    def skip_next_frame(self):
        self.__lock.acquire(True)

    def get_next_frame(self) -> Optional[bytes]:
        # when video file hasn't been passed
        if not self.video_capture:
            return

        if self.video_capture and self.video_capture.isOpened():
            grabbed, frame = self.video_capture.read()
            if not grabbed or frame is None:
                self._on_end_callback(self.source)
                if self.repeat:
                    self.video_capture.set(cv2.CAP_PROP_POS_FRAMES, 1)
                    return
                else:
                    self.stop()
                    return

            self.__pts = self.video_capture.get(cv2.CAP_PROP_POS_MSEC)

            if self.__lock.locked():
                self.__lock.release()
                return

            rgba = cv2.cvtColor(frame, cv2.COLOR_BGR2RGBA)
            rgba_bytes = rgba.tobytes()

            return rgba_bytes

    def _update(self):
        while True:
            if not self.is_running:
                break

            frame = self.get_next_frame()
            # if it was rewind
            if frame is None:
                continue

            self.__last_frame = frame
            # its necessary for thread blocking
            self.put(frame)

        if self.video_capture:
            self.video_capture.release()


class AudioStream(QueueStream):
    __REQUESTED_AUDIO_BYTES_LENGTH = DEFAULT_REQUESTED_AUDIO_BYTES_LENGTH

    def __init__(
        self,
        source: str,
        repeat: bool,
        on_end_callback,
        queue_size=AUDIO_QUEUE_SIZE,
        video_stream: Optional[VideoStream] = None,
    ):
        super().__init__(on_end_callback, queue_size)

        self.source = source
        self.__input_container = av.open(source)
        if not len(self.__input_container.streams.audio):
            raise RuntimeError('Cant find audio stream')

        codec_context = self.__input_container.streams.audio[0].codec_context

        self.__input_container.no_buffer = True
        self.__input_container.flush_packets = True
        codec_context.low_delay = True

        # TODO so strange dirty fix
        audio_rate = codec_context.sample_rate
        rate = REVERSED_AUDIO_SAMPLE_RATE.get(audio_rate, DEFAULT_AUDIO_SAMPLE_RATE)

        self.__audio_resampler = av.AudioResampler(format='s16', layout='mono', rate=rate)
        self.__audio_stream_iter = iter(self.__input_container.decode(audio=0))

        self.__video_stream = video_stream

        self.repeat = repeat

        self.__pts = 0
        self.__pts_offset = None

        self.__frame_tail = b''

    def read(self, length: int):
        self.__REQUESTED_AUDIO_BYTES_LENGTH = length

        try:
            if self.is_paused:
                raise Empty

            return super().read()
        except Empty:
            pass

    def get_pts(self):
        return self.__pts - self.__pts_offset

    def get_next_frame(self) -> Optional[List[bytes]]:
        try:
            frame = next(self.__audio_stream_iter)

            if self.__pts_offset is None:
                self.__pts_offset = frame.time * MILLIS_IN_SEC
            self.__pts = frame.time * MILLIS_IN_SEC

            if self.__video_stream:
                pts_diff = self.__video_stream.get_pts() - self.get_pts()
                if os.environ.get('DEBUG'):
                    if pts_diff < 0:
                        logger.debug(f'Video behind the audio {-pts_diff}')
                    if pts_diff > 0:
                        logger.debug(f'Audio behind the video {pts_diff}')

                # welcome to my magic values
                if pts_diff > 100:
                    logger.debug(f'Skip 1 audio frame')
                    frame = next(self.__audio_stream_iter)
                if pts_diff < -100:
                    logger.debug('Skip 1 video frame')
                    self.__video_stream.skip_next_frame()

            frame.pts = None
        except StopIteration:
            self._on_end_callback(self.source)
            if self.repeat:
                # TODO it doesnt work with live streams and will crash on the end of stream
                self.__input_container.seek(0)
                self.__audio_stream_iter = iter(self.__input_container.decode(audio=0))
                self.__frame_tail = b''
                return
            else:
                self.stop()
                return

        resampled_frame = self.__audio_resampler.resample(frame)

        frame_bytes = self.__frame_tail
        if resampled_frame:
            if isinstance(resampled_frame, list):
                # for av 9.0+
                frame_bytes += resampled_frame[0].to_ndarray().tobytes()
            else:
                # for av 8
                frame_bytes += resampled_frame.to_ndarray().tobytes()

        cut_frames = [
            frame_bytes[i : i + self.__REQUESTED_AUDIO_BYTES_LENGTH]
            for i in range(0, len(frame_bytes), self.__REQUESTED_AUDIO_BYTES_LENGTH)
        ]

        frames_to_return = []
        for frame in cut_frames:
            if len(frame) == self.__REQUESTED_AUDIO_BYTES_LENGTH:
                frames_to_return.append(frame)
            else:
                self.__frame_tail = frame

        return frames_to_return

    def _update(self):
        while True:
            if not self.is_running:
                break

            frame_list = self.get_next_frame()
            if frame_list is None:
                continue

            for frame in frame_list:
                self.put(frame)

        self.__input_container.close()



from typing import Callable

import tgcalls
from pytgcalls.implementation import GroupCallBase
from pytgcalls.utils import DEFAULT_COLOR_DEPTH, VideoInfo


class GroupCallRaw(GroupCallBase):
    def __init__(
        self,
        mtproto_bridge,
        on_audio_played_data: Callable[['GroupCallRaw', int], bytes] = None,
        on_audio_recorded_data: Callable[['GroupCallRaw', bytes, int], None] = None,
        on_video_played_data: Callable[['GroupCallRaw'], bytes] = None,
        enable_logs_to_console=False,
        path_to_log_file=None,
        outgoing_audio_bitrate_kbit=128,
    ):
        super().__init__(mtproto_bridge, enable_logs_to_console, path_to_log_file, outgoing_audio_bitrate_kbit)

        self.__is_playout_paused = False
        self.__is_recording_paused = False

        self.__raw_audio_device_descriptor = None

        self.on_audio_played_data = on_audio_played_data
        self.on_audio_recorded_data = on_audio_recorded_data

        self.on_video_played_data = on_video_played_data

    def __create_and_return_raw_audio_device_descriptor(self):
        self.__raw_audio_device_descriptor = tgcalls.RawAudioDeviceDescriptor()
        self.__raw_audio_device_descriptor.getPlayedBufferCallback = self.__get_played_audio_buffer_callback
        self.__raw_audio_device_descriptor.setRecordedBufferCallback = self.__set_recorded_audio_buffer_callback
        self.__raw_audio_device_descriptor.isPlayoutPaused = self.__is_playout_paused_callback
        self.__raw_audio_device_descriptor.isRecordingPaused = self.__is_recording_paused_callback

        return self.__raw_audio_device_descriptor

    def _setup_and_start_group_call(self):
        self._start_native_group_call(self.__create_and_return_raw_audio_device_descriptor())
        self._configure_video_capture(VideoInfo.default())

    def _configure_video_capture(self, video_info: VideoInfo):
        self._set_video_capture(
            self.__get_played_video_buffer_callback, video_info.width, video_info.height, video_info.fps
        )

    def __get_played_video_buffer_callback(self):
        frame = b''
        if self.on_video_played_data:
            data = self.on_video_played_data(self)
            if data:
                frame = data

        video_info = VideoInfo.default()
        if self._video_stream:
            video_info = self._video_stream.get_video_info()

        frame_size = video_info.width * video_info.width * DEFAULT_COLOR_DEPTH
        return frame.ljust(frame_size, b'\0')

    def __get_played_audio_buffer_callback(self, length: int):
        frame = b''
        if self.on_audio_played_data:
            data = self.on_audio_played_data(self, length)
            if data:
                frame = data

        return frame.ljust(length, b'\0')

    def __set_recorded_audio_buffer_callback(self, frame: bytes, length: int):
        if self.on_audio_recorded_data:
            self.on_audio_recorded_data(self, frame, length)

    def __is_playout_paused_callback(self):
        # native pausing from cpp side. old way to impl it
        return self.__is_playout_paused

    def __is_recording_paused_callback(self):
        # native pausing from cpp side. old way to impl it
        return self.__is_recording_paused



from .base_wrapper import WrapperBase
from .group_call_wrapper import GroupCallWrapper
from .group_call_discarded_wrapper import GroupCallDiscardedWrapper
from .group_call_participant_wrapper import GroupCallParticipantWrapper

__all__ = [
    'WrapperBase',
    'GroupCallWrapper',
    'GroupCallDiscardedWrapper',
    'GroupCallParticipantWrapper',
]



from typing import List, TYPE_CHECKING

from pytgcalls.mtproto.data import WrapperBase

if TYPE_CHECKING:
    from pytgcalls.mtproto.data import GroupCallParticipantWrapper


class UpdateGroupCallParticipantsWrapper(WrapperBase):
    def __init__(self, participants: List['GroupCallParticipantWrapper']):
        self.participants = participants



from setuptools import setup, find_packages

# Membaca isi dari requirements.txt
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='geramcals',
    version='1.0',
    packages=find_packages(),
    description='Seratus Persen Clone Pytgcalls',
    author='pyrogeram',
    author_email='pyrogeram@gmail.com',
    url='https://t.me/pypyrogramcalls',  # Ganti dengan URL proyek Anda
    classifiers=[
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=requirements,
)




from asyncio import AbstractEventLoop
from typing import Callable

from telethon.errors import (
    BadRequestError as TelethonBadRequestError,
    GroupcallJoinMissingError as TelethonGroupcallJoinMissingError,
    GroupcallSsrcDuplicateMuchError as TelethonGroupcallSsrcDuplicateMuchError,
)
from telethon.events import Raw, StopPropagation
from telethon.tl import functions
from telethon.tl.types import (
    DataJSON,
    GroupCallDiscarded as TelethonGroupCallDiscarded,
    SpeakingInGroupCallAction,
    UpdateGroupCall,
    UpdateGroupCallConnection,
    UpdateGroupCallParticipants,
    InputPeerChat,
    InputPeerChannel,
)

from pytgcalls import PytgcallsError
from pytgcalls.mtproto import MTProtoBridgeBase
from pytgcalls.mtproto.data import GroupCallDiscardedWrapper, GroupCallParticipantWrapper, GroupCallWrapper
from pytgcalls.mtproto.data.update import UpdateGroupCallParticipantsWrapper, UpdateGroupCallWrapper
from pytgcalls.mtproto.exceptions import BadRequest, GroupcallSsrcDuplicateMuch
from pytgcalls.utils import int_ssrc


class TelethonBridge(MTProtoBridgeBase):
    def __init__(self, client):
        super().__init__(client)

        self._loop = client.loop

        self._update_to_handler = {
            UpdateGroupCallParticipants: self._process_group_call_participants_update,
            UpdateGroupCall: self._process_group_call_update,
        }

    async def _process_update(self, update):
        if type(update) not in self._update_to_handler.keys():
            return

        if not self.group_call or not update.call or update.call.id != self.group_call.id:
            return
        self.group_call = update.call

        await self._update_to_handler[type(update)](update)
        raise StopPropagation

    async def _process_group_call_participants_update(self, update):
        participants = [GroupCallParticipantWrapper.create(p) for p in update.participants]
        wrapped_update = UpdateGroupCallParticipantsWrapper(participants)

        await self.group_call_participants_update_callback(wrapped_update)

    async def _process_group_call_update(self, update):
        if not isinstance(update.call, TelethonGroupCallDiscarded):
            return

        call = GroupCallDiscardedWrapper()  # no info needed
        wrapped_update = UpdateGroupCallWrapper(update.chat_id, call)

        await self.group_call_update_callback(wrapped_update)

    async def _process_group_call_connection(self, update):
        # TODO update to new layer when pyrogram will release new stable version on pypi
        call = GroupCallWrapper('placeholder', update.params)
        wrapped_update = UpdateGroupCallWrapper('placeholder', call)

        await self.group_call_update_callback(wrapped_update)

    async def check_group_call(self) -> bool:
        if not self.group_call or not self.my_ssrc:
            return False

        try:
            ssrcs_in_group_call = await (
                self.client(
                    functions.phone.CheckGroupCallRequest(call=self.group_call, sources=[int_ssrc(self.my_ssrc)])
                )
            )
        except TelethonBadRequestError as e:
            if isinstance(e, TelethonGroupcallJoinMissingError):
                return False
            else:
                raise BadRequest(e)

        return int_ssrc(self.my_ssrc) in ssrcs_in_group_call

    async def leave_current_group_call(self):
        if not self.full_chat or not self.full_chat.call or not self.my_ssrc:
            return

        response = await self.client(
            functions.phone.LeaveGroupCallRequest(call=self.full_chat.call, source=int_ssrc(self.my_ssrc))
        )

        await self.handle_updates(response)

    async def edit_group_call_member(
        self, peer, volume: int = None, muted=False, video_stopped=True, video_paused=False
    ):
        response = await self.client(
            functions.phone.EditGroupCallParticipantRequest(
                call=self.full_chat.call,
                participant=peer,
                muted=muted,
                volume=volume,
                video_stopped=video_stopped,
                video_paused=video_paused,
            )
        )

        await self.handle_updates(response)

    async def get_and_set_self_peer(self):
        self.my_peer = await self.client.get_me(input_peer=True)

        return self.my_peer

    async def get_and_set_group_call(self, group):
        self.chat_peer = group

        if type(group) not in [InputPeerChannel, InputPeerChat]:
            self.chat_peer = await self.client.get_input_entity(group)

        if isinstance(self.chat_peer, InputPeerChannel):
            self.full_chat = (await self.client(functions.channels.GetFullChannelRequest(group))).full_chat
        elif isinstance(self.chat_peer, InputPeerChat):
            self.full_chat = (await self.client(functions.messages.GetFullChatRequest(group))).full_chat

        if self.full_chat is None:
            raise PytgcallsError(f'Can\'t get full chat by {group}')

        self.group_call = self.full_chat.call

        return self.group_call

    def unregister_update_handlers(self):
        self.client.remove_event_handler(self._process_update, Raw)

    def register_update_handlers(self):
        self.client.add_event_handler(self._process_update, Raw)

    async def resolve_and_set_join_as(self, join_as):
        if join_as is None:
            self.join_as = self.full_chat.groupcall_default_join_as
            if self.join_as is None:
                self.join_as = await self.get_and_set_self_peer()
        else:
            self.join_as = join_as

    async def send_speaking_group_call_action(self):
        await self.client(functions.messages.SetTypingRequest(peer=self.chat_peer, action=SpeakingInGroupCallAction()))

    async def join_group_call(
        self, invite_hash: str, params: str, muted: bool, video_stopped: bool, pre_update_processing: Callable
    ):
        try:
            response = await self.client(
                functions.phone.JoinGroupCallRequest(
                    call=self.group_call,
                    join_as=self.join_as,
                    invite_hash=invite_hash,
                    params=DataJSON(data=params),
                    muted=muted,
                    video_stopped=video_stopped,
                )
            )

            pre_update_processing()

            # it is here cuz we need to associate params for connection with group call
            for update in response.updates:
                if isinstance(update, UpdateGroupCallConnection):
                    await self._process_group_call_connection(update)

            await self.handle_updates(response)
        except TelethonGroupcallSsrcDuplicateMuchError as e:
            raise GroupcallSsrcDuplicateMuch(e)

    async def handle_updates(self, updates):
        for update in updates.updates:
            try:
                await self._process_update(update)
            except StopPropagation:
                break
